'use strict';

var _             = require('underscore')
  , anchor        = require('anchor-markdown-header')
  , updateSection = require('update-section')
  , getHtmlHeaders = require('./get-html-headers');

function matchesStart(line) {
  return (/<!-- START doctoc generated TOC /).test(line);
}

function matchesEnd(line) {
  return (/<!-- END doctoc generated TOC /).test(line);
}

function notNull(x) { return  x !== null; }

function addAnchor(mode, header) {
  var regExp = /\(#([^)]+)\)/;
  var anchorParse = anchor(header.name, mode, header.instance);
  var matches = regExp.exec(anchorParse);

  header.anchor = anchorParse;
  header.rawAnchor = matches[1];
  return header;
}


function getHashedHeaders (lines) {
  var inCodeBlock = false
    , lineno = 0
    , origin = []
    , preLine = 0
    , total = lines.length;

  // Turn all headers into '## xxx' even if they were '## xxx ##'
  function normalize(header) {
    return header.replace(/[ #]+$/, '');
  }

  // Find headers of the form '### xxxx xxx xx [###]'
  return lines
    .map(function (x, idx) {
      var linesObj = { lineno: idx, line: x }
      origin.push(linesObj);
      return linesObj;
    })
    .filter(function (x) {
      if (x.line.match(/^```/)) {
        inCodeBlock = !inCodeBlock;
      }
      return !inCodeBlock;
    })
    .map(function (x, idx) {
      var match = /^(\#{1,8})[ ]*(.+)\r?$/.exec(x.line);
      
      if(match){
        // first title match
        preLine = x.lineno;
        var contentArr = [];
        var blockMode = false;

        var last = origin
          .filter(function(x) {
            var range = _.range(preLine + 1, total);
            if(range.indexOf(x.lineno) !== -1) 
              return x;
          })
        var nextLine = _.find(last, function(y) {

            var match = /^(\#{1,8})[ ]*(.+)\r?$/.exec(y.line);
            
            if(!blockMode) {
              // not in block
              if(y.line.match(/^```/)) {
                blockMode = true;
              }
            }else {
              // in block
              if(y.line.match(/^```/)) {
                blockMode = false;
              }
            }

            if(match) {
              if(!blockMode)
                return contentArr;
              else
                contentArr.push(y);
            }else {
              contentArr.push(y)
            }
          })
      }

      return match 
        ? { rank :  match[1].length
          , name :  normalize(match[2])
          , line :  x.lineno
          , content : contentArr.map(function(x) {
            return x.line;
          }).join('\n')

          }
        : null;
    })
    .filter(notNull)
}

function getUnderlinedHeaders (lines) {
    // Find headers of the form
    // h1       h2
    // ==       --

    var origin = []
      , preLine = 0
      , total = lines.length;

    return lines
      .map(function (line, index, lines_) {
        if (index === 0) return null;
        var rank;
        var match = false;

        if (/^==+ *\r?$/.exec(line)){
          rank = 1;
          match = true;
        }else if (/^--+ *\r?$/.exec(line)){
          rank = 2;
          match = true;
        }else{
          match = false;
          return null;
        }

        if(lines[index -1 ].trim() === '') {
          match = false;
          return null;
        }

        if(match){
          // first title match
          preLine = index;
          var contentArr = [];
          var blockMode = false;

          var last = lines
            .filter(function(x, idx) {
              var range = _.range(preLine + 1, total);
              if(range.indexOf(idx) !== -1) 
                return x;
            })

          var nextLine = _.find(last, function(y) {

              var match = false;
              if (/^==+ *\r?$/.exec(y)){
                match = true;
              }else if (/^--+ *\r?$/.exec(y)){
                match = true;
              }

              if(!blockMode) {
                // not in block
                if(y.match(/^```/)) {
                  blockMode = true;
                }
              }else {
                // in block
                if(y.match(/^```/)) {
                  blockMode = false;
                }
              }

              if(match) {
                if(!blockMode){
                  contentArr.pop();
                  return contentArr;
                }else{
                  contentArr.push(y);
                }                 
              }else {
                contentArr.push(y)
              }
            })
        }

        return {
          rank  :  rank,
          name  :  lines_[index - 1],
          line  :  index - 1, 
          content : contentArr.join('\n')
        };
      })
      .filter(notNull)
}

function countHeaders (headers) {
  var instances = {};

  for (var i = 0; i < headers.length; i++) {
    var header = headers[i];
    var name = header.name;

    if (instances.hasOwnProperty(name)) {
      instances[name]++;
    } else {
      instances[name] = 0;
    }

    header.instance = instances[name];
  }

  return headers;
}

function getLinesToToc (lines, currentToc, info) {
  if (!currentToc) return lines;

  var tocableStart = 0;

  // when updating an existing toc, we only take the headers into account
  // that are below the existing toc
  if (info.hasEnd) tocableStart = info.endIdx;

  return lines.slice(tocableStart);
}

exports = module.exports = function transform(content, maxHeaderNo, mode, cb) {
  if(arguments.length === 3 && isNaN(arguments[1])) {
    cb = mode;
    mode = maxHeaderNo;
    maxHeaderNo = null;
  }else if(arguments.length === 3 && !isNaN(arguments[1])) {
    cb = mode;
    mode = null;
  }else if(arguments.length === 2) {
    cb = maxHeaderNo;
    maxHeaderNo = null;
  }

  mode = mode || 'github.com';
  maxHeaderNo = maxHeaderNo || 4;

  var lines = content.split('\n')
    , info = updateSection.parse(lines, matchesStart, matchesEnd)

  var currentToc = info.hasStart && lines.slice(info.startIdx, info.endIdx).join('\n')
    , linesToToc = getLinesToToc(lines, currentToc, info);

  var headers = getHashedHeaders(linesToToc)
    .concat(getUnderlinedHeaders(linesToToc))
    .concat(getHtmlHeaders(linesToToc, maxHeaderNo))

  headers.sort(function (a, b) {
    return a.line - b.line;
  });

  var allHeaders    =  countHeaders(headers)
    , lowestRank    =  _(allHeaders).chain().pluck('rank').min().value()
    , linkedHeaders =  _(allHeaders).map(addAnchor.bind(null, mode));

  if (linkedHeaders.length === 0) return { transformed: false };

  var toc =
    linkedHeaders
        .map(function (x) {
          var indent = _(_.range(x.rank - lowestRank))
            .reduce(function (acc, x) { return acc + '  '; }, '');

          return indent + '- ' + x.anchor;
        })
        .join('\n')
    + '\n';

  if (currentToc === toc) return cb(null);

  cb({content: headers, toc: toc});
};

